#include "resource.h"

#include <ddraw.h>
#include <windows.h>
#include <sys/timeb.h>
#include <stdio.h>
#include <fcntl.h>
#include <io.h>
#include "version.h"
#include "prototypes.h"

#define FGCOLOR RGB(0, 0, 0)
#define BGCOLOR RGB(255,255,255)

HWND	hWnd = NULL,
		hWndDebug = NULL;
MSG		msg;

#define	SCREEN_X	196
#define SCREEN_Y	132
#define	RELEASE(s)	if(s){IDirectDrawSurface_Release(s); s = NULL; }
RECT				rcScreen;
LPDIRECTDRAW		lpDD;
LPDIRECTDRAWSURFACE	sPrimary,
					sBackBuffer;
DWORD				dwScreen[48][32];
BOOL				Running = FALSE,
					DebugVis = FALSE;

// Get the coords of the rendering area
void DDReSize(){
	GetClientRect(hWnd, &rcScreen);
	ClientToScreen(hWnd, (POINT*)&rcScreen.left);
	ClientToScreen(hWnd, (POINT*)&rcScreen.right);
}

void DDClear(){
	struct _DDBLTFX ddbltfx;
	ZeroMemory(&ddbltfx, sizeof(ddbltfx));
	ddbltfx.dwSize  = sizeof(ddbltfx);
	ddbltfx.dwFillColor = BGCOLOR;
	IDirectDrawSurface_Blt(sBackBuffer, NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
}

void DDShutdown(){
	if(	lpDD){
		RELEASE(sBackBuffer);
		RELEASE(sPrimary);
		IDirectDraw_Release(lpDD);
		lpDD = NULL;
	}
}

void DDRestore(){
	IDirectDrawSurface_Restore(sPrimary);
	IDirectDrawSurface_Restore(sBackBuffer);
}


void DDInit(){
	DDSCAPS2			ddscaps;	// ddsurface capabilities struct
	DDSURFACEDESC		ddsd;		// ddsurface description struct
    LPDIRECTDRAWCLIPPER pClipper;

	DDShutdown();

	ZeroMemory(&ddscaps, sizeof(ddscaps) );
	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);

	DirectDrawCreate(NULL, &lpDD, NULL);
	IDirectDraw_SetCooperativeLevel(lpDD, hWnd, DDSCL_NORMAL);

	DDReSize();

	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_SYSTEMMEMORY;

	IDirectDraw_CreateSurface(lpDD, &ddsd, &sPrimary, NULL);
	IDirectDraw_CreateClipper(lpDD, 0, &pClipper, NULL);
	IDirectDrawClipper_SetHWnd(pClipper, 0, hWnd);
	IDirectDrawSurface_SetClipper(sPrimary, pClipper);
	IDirectDrawClipper_Release(pClipper);
	pClipper = NULL;

	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.dwWidth = SCREEN_X;
	ddsd.dwHeight= SCREEN_Y;
	ddsd.ddsCaps.dwCaps = DDSCAPS_SYSTEMMEMORY;
	IDirectDraw_CreateSurface(lpDD, &ddsd, &sBackBuffer,NULL);

	Running = TRUE;
}

void low_redrawlcd()
{
	IDirectDrawSurface_Blt(sPrimary, &rcScreen, sBackBuffer, NULL, DDBLT_WAIT, NULL);
}

void redrawlcd()
{
	low_redrawlcd();
}

void SyncDebugVis(){
	ShowWindow(hWndDebug, (DebugVis ? SW_SHOW : SW_HIDE));
	CheckMenuItem(GetMenu(hWnd), ID_VIEW_DEBUGWINDOW, (DebugVis ? MF_CHECKED : MF_UNCHECKED));
}

DWORD DecToBin(DWORD v){
	DWORD r;
	r = v & 1;
	r += ((v>>1) & 1) * 10;
	r += ((v>>2) & 1) * 100;
	r += ((v>>3) & 1) * 1000;
	r += ((v>>4) & 1) * 10000;
	r += ((v>>5) & 1) * 100000;
	r += ((v>>6) & 1) * 1000000;
	r += ((v>>7) & 1) * 10000000;
	return r;
}

extern unsigned char sfr[0x100];

void UpdateDebugValues(){
	char t[100];

	if(!DebugVis) return;

	sprintf(t, "%03d", sfr[0]);
	SetDlgItemText(hWndDebug, IDC_EDIT_ACC, t);
	sprintf(t, "%02X", sfr[0]);
	SetDlgItemText(hWndDebug, IDC_EDIT_ACC2, t);
	sprintf(t, "%08d", DecToBin(sfr[0]));
	SetDlgItemText(hWndDebug, IDC_EDIT_ACC3, t);

	sprintf(t, "%03d", sfr[2]);
	SetDlgItemText(hWndDebug, IDC_EDIT_B, t);
	sprintf(t, "%02X", sfr[2]);
	SetDlgItemText(hWndDebug, IDC_EDIT_B2, t);
	sprintf(t, "%08d", DecToBin(sfr[2]));
	SetDlgItemText(hWndDebug, IDC_EDIT_B3, t);

	sprintf(t, "%03d", sfr[3]);
	SetDlgItemText(hWndDebug, IDC_EDIT_C, t);
	sprintf(t, "%02X", sfr[3]);
	SetDlgItemText(hWndDebug, IDC_EDIT_C2, t);
	sprintf(t, "%08d", DecToBin(sfr[3]));
	SetDlgItemText(hWndDebug, IDC_EDIT_C3, t);
}

BOOL CALLBACK DebugProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam){

	switch(msg) {
	case WM_COMMAND:
		switch(LOWORD(wparam))
		{
		case IDC_BUTTON_UPDATE:
			UpdateDebugValues();
			break;
		case IDCANCEL:
			DebugVis = FALSE;
			SyncDebugVis();
			break;
		}
	break;
	}
	return FALSE;
}

extern BOOL CALLBACK DownloadProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);


// MAIN WINDOWS' CALLBACK FUNCTION //////////////////////////////////////////////////
LRESULT CALLBACK MainWndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam){
	static BOOL	lastwparam = TRUE;
	WORD		fActive;

	switch(msg) {
		case WM_COMMAND:
			switch( LOWORD(wparam) )
            {
			case ID_FILE_LOAD: {
				char *	fn,
						hdr[4];
				int		fd,
						r;

				fn = ShowOpenDialog(GetWindowLong(hwnd, GWL_HINSTANCE));
				if(!fn)	break;

				fd = open(fn, O_RDONLY);
				if(fd<0 || (r=read(fd, hdr, 4)<4)){
					close(fd);
					break;
				}

				if(hdr[0]=='L' && hdr[1]=='C' && hdr[2]=='D' && hdr[3]=='i')
						do_lcdimg(fn);
				else	do_vmsgame(fn, NULL);
			   } break;
			case ID_FILE_DOWNLOAD:
				DialogBoxParam(	(HINSTANCE)GetWindowLong(hwnd, GWL_HINSTANCE),
								MAKEINTRESOURCE(IDD_DOWNLOAD),
								hwnd,
								(DLGPROC)DownloadProc,
								NULL);
				break;
			case ID_FILE_RESET:
				resetcpu();
				break;
			case ID_FILE_QUIT:
				PostQuitMessage(0);
				break;
			case ID_VIEW_DEBUGWINDOW:
				DebugVis = !DebugVis;
				SyncDebugVis();
				break;
			case ID_VIEW_RESTOREDEFAULTSIZE:
  				SetWindowPos(	hWnd, HWND_NOTOPMOST, 0, 0,
								SCREEN_X+(GetSystemMetrics(SM_CXSIZEFRAME)*2),
								SCREEN_Y+(GetSystemMetrics(SM_CYSIZEFRAME)*2)+(GetSystemMetrics(SM_CYMENU)*2),
								SWP_NOMOVE);
				DDReSize();
				break;
			case ID_HELP_KEYS: {
				char message[1000];
				sprintf(message, "%s%s%s%s%s%s",
						"Escape - Exit program.\n",
						"Arrow Keys - Directional pad.\n",
						"A/Z - 'A' button.\n",
						"B/X - 'B' button.\n",
						"M/Return - Mode button.\n",
						"S/Space - Sleep button.\n");
				MessageBox(NULL, message, "Controls", MB_OK);
			} break;
			case ID_HELP_ABOUT: {
				char message[1000];
				sprintf(message, "%s",
						"DirectX port by biggs\nhttp://rpging.simplenet.com");
				MessageBox(NULL, message, "About", MB_OK);
			} break;
		}
		break;
		case WM_ACTIVATEAPP:
		{
			fActive = (BOOL)wparam;           // activation flag
			if (fActive) {
				if (fActive != lastwparam) {
					DDRestore();
					Running = TRUE;
				}
			} else {
				Running = FALSE;
			}

			lastwparam = fActive;
			return 0;
		} break;

		case WM_SIZE:
        case WM_MOVE:
			DDReSize(hwnd);
			break;

		case WM_DESTROY:
			PostQuitMessage(0);
			break;

		case WM_PAINT:
			if(Running)
				low_redrawlcd();
			break;

		case WM_KEYDOWN:
			switch(wparam & 0xff) {
			case VK_ESCAPE:	PostQuitMessage(0);		return 0;
			case VK_UP:		keypress(0);			return 0;
			case VK_DOWN:	keypress(1);			return 0;
			case VK_LEFT:	keypress(2);			return 0;
			case VK_RIGHT:	keypress(3);			return 0;
			case 'Z':
			case 'A':		keypress(4);			return 0;
			case 'X':
			case 'B':		keypress(5);			return 0;
			case VK_RETURN:
			case 'M':		keypress(6);			return 0;
			case VK_SPACE:
			case 'S':		keypress(7);			return 0;
			case 'D':		UpdateDebugValues();	return 0;
			}
			break;
		case WM_KEYUP:
			switch(wparam & 0xff) {
			case VK_UP:		keyrelease(0);			return 0;
			case VK_DOWN:	keyrelease(1);			return 0;
			case VK_LEFT:	keyrelease(2);			return 0;
			case VK_RIGHT:	keyrelease(3);			return 0;
			case 'Z':
			case 'A':		keyrelease(4);			return 0;
			case 'X':
			case 'B':		keyrelease(5);			return 0;
			case VK_RETURN:
			case 'M':		keyrelease(6);			return 0;
			case VK_SPACE:
			case 'S':		keyrelease(7);			return 0;
			}
			break;
	} // end switch message

	// let windows process any messages that we didn't take care of 
	return (DefWindowProc(hwnd, msg, wparam, lparam));
} // end WinProc


WNDCLASS wnd_class = {
  0,
  MainWndProc,
  0, 0,
  NULL,
  NULL,
  NULL,
  (HBRUSH)COLOR_BTNFACE,
  NULL,
  "softvms_wnd_class"
};


int openmainwin(HINSTANCE inst)
{  
  wnd_class.hInstance = inst;
  if(!RegisterClass(&wnd_class))
    return 0;

  hWnd = CreateWindow("softvms_wnd_class", "SoftDirectVMS " VERSION, WS_OVERLAPPEDWINDOW,
						CW_USEDEFAULT, CW_USEDEFAULT,
  						SCREEN_X+(GetSystemMetrics(SM_CXSIZEFRAME)*2),
						SCREEN_Y+(GetSystemMetrics(SM_CYSIZEFRAME)*2)+(GetSystemMetrics(SM_CYMENU)*2),
						NULL,	//parent
						LoadMenu(inst, MAKEINTRESOURCE(IDR_MENU_MAIN)),
						inst,
						NULL);

  if(!hWnd)
    return 0;

  UpdateWindow(hWnd);
  ShowWindow(hWnd, SW_SHOW);

  hWndDebug = CreateDialog(	inst,
							MAKEINTRESOURCE(IDD_DEBUG),
							hWnd,
							(DLGPROC)DebugProc);

	if(!hWndDebug)
		return 0;

  UpdateWindow(hWndDebug);

  DDInit();

  return 1;
}

void deinitmainwin(HINSTANCE inst)
{
	DDShutdown();
	DestroyWindow(hWnd);
	DestroyWindow(hWndDebug);
	UnregisterClass("softvms_wnd_class", inst);
}


void vmputpixel(int x, int y, int p)
{
	if(p&1){
		RECT drect;
		struct _DDBLTFX ddbltfx;
		SetRect(&drect, (x << 2) + 1, (y << 2) + 1, (x + 1) << 2, (y + 1) << 2);
		ZeroMemory(&ddbltfx, sizeof(ddbltfx));
		ddbltfx.dwSize  = sizeof(ddbltfx);
		ddbltfx.dwFillColor = FGCOLOR;
		IDirectDrawSurface_Blt(sBackBuffer, &drect, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
	}
}

void checkevents() {
  while(PeekMessage(&msg, (HWND) NULL, 0, 0, PM_REMOVE)) {
    if(msg.message == WM_QUIT) {
      deinitmainwin(wnd_class.hInstance);
      exit(0);
    }
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
}

void waitforevents(struct timeval *tv)
{
  DWORD millis = INFINITE;
  if(tv != NULL)
    millis = tv->tv_sec * 1000 + tv->tv_usec / 1000;
  MsgWaitForMultipleObjects(0, NULL, FALSE, millis, QS_ALLINPUT);
}

/* UNIX compat... */

int gettimeofday(struct timeval *tv)
{
  struct _timeb timebuffer;
  _ftime( &timebuffer );
  tv->tv_sec = timebuffer.time;
  tv->tv_usec = timebuffer.millitm*1000;
  return 0;
}

void sleep(int t)
{
  Sleep(t*1000);
}
